"""
psycopg distribution version file.
"""

# Copyright (C) 2020 The Psycopg Team

# Use a versioning scheme as defined in
# https://www.python.org/dev/peps/pep-0440/

# STOP AND READ! if you change:
__version__ = "3.1.18"
# also change:
# - `docs/news.rst` to declare this as the current version or an unreleased one
# - `psycopg_c/psycopg_c/version.py` to the same version.
